import sys
from typing import Dict, List, NoReturn, Optional, Tuple, TypedDict

from .error import LexerError
from .Token import Token, TokenType

RESERVED_KEYWORDS: Dict[str, TokenType] = {
    t.value: t
    for t in list(TokenType)
    if t.value.isalpha() and t.value not in ["name", "number", "eof", "string"]
}
SYMBOLS: Dict[str, TokenType] = {
    t.value: t for t in list(TokenType) if not t.value.isalpha()
}
# string.isnumeric works on unicode numbers, too. lua only works on Arabic-Indic digits
NUMBER: List[str] = [str(i) for i in range(10)]
HEX_NUMBER: List[str] = [
    *(chr(i + 65) for i in range(6)),
    *(chr(i + 97) for i in range(6)),
    *NUMBER,
]
LETTER: List[str] = [
    *(chr(i + 65) for i in range(26)),
    *(chr(i + 97) for i in range(26)),
    "_",
]
ALPHANUMERIC: List[str] = [*NUMBER, *LETTER]
ESCAPE_CODES: Dict[str, str] = {
    "a": "\a",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "\n": "\n",
    "r": "\r",
    "t": "\t",
    "v": "\v",
    "\\": "\\",
    '"': '"',
    "'": "'",
}

NumberTuple = Tuple[bool, Optional[str], Optional[str], Optional[str], Optional[str]]


class TokenArgs(TypedDict):
    line: int
    column: int
    comment: list[str]


def include_typing(do_include: bool) -> None:
    """
    Change the keywords to include typing

    :param do_include: whether to include typing keywords
    :return: None
    """
    if do_include and "as" not in RESERVED_KEYWORDS:
        RESERVED_KEYWORDS["as"] = TokenType.AS
        RESERVED_KEYWORDS["is"] = TokenType.IS
    elif not do_include and "as" in RESERVED_KEYWORDS:
        RESERVED_KEYWORDS.pop("as")
        RESERVED_KEYWORDS.pop("is")


class Lexer:
    def __init__(
        self, text: str, typed: bool = False, ignore_unicode_errors: bool = False
    ) -> None:
        self.text: str = text
        self.text_len: int = len(self.text)
        self.line: int = 0
        self.column: int = -1
        self.pos: int = -1
        self.newline_warn: int = 0
        self.current_char: Optional[str] = None
        self.last_hint: Optional[Tuple[str, int, int]] = None
        self.comments: list[str] = []
        self.unicode_errors: str = "ignore" if ignore_unicode_errors else "strict"
        self.advance()
        include_typing(typed)

    def error(
        self, message: str, line: Optional[int] = None, column: Optional[int] = None
    ) -> NoReturn:
        current_line: int = line if line is not None else self.line
        current_column = column if column is not None else self.column
        print(f"Error on line {current_line + 1}:", file=sys.stderr)
        print(self.text.split("\n")[current_line], file=sys.stderr)
        print(" " * current_column + "^", file=sys.stderr)
        print(message, file=sys.stderr)
        raise LexerError(message, current_line, current_column)

    def get_token_args(self) -> TokenArgs:
        comment: list[str] = self.comments[:]
        self.comments.clear()
        return {"line": self.line + 1, "column": self.column + 1, "comment": comment}

    def advance(self) -> None:
        """Advance the `pos` pointer"""
        self.pos += 1
        if self.pos < self.text_len:
            self.current_char = self.text[self.pos]
            self.column += 1

            if self.current_char == "\n":
                self.column = -1
                self.line += 1
                # init newline state machine
                self.newline_warn = 1
            elif self.current_char.isspace():
                # third state: space after text after newline
                if self.newline_warn == 2:
                    self.newline_warn = 3
            else:
                # second state: text after newline
                if self.newline_warn == 1:
                    self.newline_warn = 2
                # reset if text after third state
                if self.newline_warn == 3:
                    self.newline_warn = 0
        else:
            self.current_char = None

    def peek(self) -> Optional[str]:
        peek_pos = self.pos + 1
        if peek_pos < self.text_len:
            return self.text[peek_pos]
        return None

    def skip_whitespace(self) -> None:
        while self.current_char and self.current_char.isspace():
            self.advance()

    def get_long_brackets(self) -> str:
        """Returns the inner content of long brackets, or none if there are no long brackets"""
        # check if in the right conditions
        assert self.current_char == "[" and self.peek() in ["=", "["]
        line: int = self.line
        column: int = self.column
        # skip opening bracket
        self.advance()
        # the amount of equals signs in the long string
        equals: int = 0
        # get all equals signs
        while self.current_char == "=":
            equals += 1
            self.advance()
        # check that the opening is not malformed
        if self.current_char != "[":
            self.error("Malformed long bracket")
        self.advance()
        inner_string: str = ""
        # closing_equals is zero on an encountered closing bracket, and will be incremented with each equals sign
        closing_equals: int = -1
        while self.current_char:
            if self.current_char == "]" and closing_equals == equals:
                self.advance()
                return inner_string[: -equals - 1]
            if self.current_char == "=" and closing_equals >= 0:
                closing_equals += 1
            elif self.current_char == "]":
                closing_equals = 0
            else:
                closing_equals = -1
            inner_string += self.current_char
            self.advance()
        self.error("long brackets never closed", line, column)

    def skip_comment(self) -> None:
        """Skip a comment (long or short)"""
        assert self.current_char == "-" and self.peek() == "-"
        self.advance()
        self.advance()
        comment: str = ""
        if self.current_char == "[" and self.peek() in ["[", "="]:
            comment = self.get_long_brackets()
        else:
            while self.current_char and self.current_char != "\n":
                comment += self.current_char
                self.advance()
        self.comments.append(comment)

    def get_number(self) -> NumberTuple:
        """Parses a number into the Number ast node"""
        current_numbers: List[str] = NUMBER
        is_hex: bool = False
        integer_part: Optional[str] = None
        # starts with a digit
        if self.current_char in NUMBER:
            # is a hex number
            if self.current_char == "0" and self.peek() in ["x", "X"]:
                is_hex = True
                current_numbers = HEX_NUMBER
                self.advance()
                self.advance()
            # gather integer digits
            result: str = ""
            while self.current_char in current_numbers:
                result += self.current_char.lower()
                self.advance()
            if result:
                integer_part = result
        if not integer_part:
            self.last_hint = (
                "forgot an integer part of a number",
                self.line,
                self.column,
            )
        # has a fractional part
        fractional_part: Optional[str] = None
        exponent: Optional[str] = None
        float_offset: Optional[str] = None
        if self.current_char == ".":
            self.advance()
            # gather fractional digits
            result = ""
            while self.current_char in current_numbers:
                result += self.current_char.lower()
                self.advance()
            if result:
                fractional_part = result
            else:
                self.last_hint = (
                    "forgot a fractional part after a dot",
                    self.line,
                    self.column,
                )
        # test for a float_offset in hex numbers or an exponent in non-hex numbers
        if (
            is_hex
            and self.current_char in ["p", "P"]
            or not is_hex
            and self.current_char in ["e", "E"]
        ):
            self.advance()
            result = ""
            # take a sign if it exists
            if self.current_char in ["+", "-"]:
                result += self.current_char
                self.advance()
            # gather digits
            while self.current_char in NUMBER:
                result += self.current_char
                self.advance()
            # store in the respective field
            if result and is_hex:
                float_offset = result
            elif result:
                exponent = result
        return is_hex, integer_part, fractional_part, exponent, float_offset

    def get_name(self) -> str:
        assert self.current_char in LETTER
        result: str = ""
        while self.current_char in ALPHANUMERIC:
            result += self.current_char
            self.advance()
        return result

    def _safe_decode(self, base: int) -> str:
        try:
            return bytes((base,)).decode("utf-8", self.unicode_errors)
        except UnicodeDecodeError:
            self.error(
                f"This library can't handle invalid unicode characters, got {base}"
            )

    def _safe_code_point(self, base: int) -> str:
        try:
            return chr(base)
        except ValueError:
            if self.unicode_errors == "strict":
                self.error(f"Invalid unicode codepoint, got {base}")
        return ""

    def get_string(self) -> str:
        assert self.current_char in ["'", '"']
        # character that is needed to close the string
        closing: str = self.current_char
        # whether the next character has been escaped
        escape: bool = False
        self.advance()
        # the result string that will contain the parsed string
        result: str = ""
        # iterate until either the end is reached, or an unescaped closing character is encountered
        while self.current_char and (escape or self.current_char != closing):
            # handle an escaped character
            if escape:
                column: int = self.column
                # handle the skip next whitespace escape sequence
                if self.current_char == "z":
                    self.advance()
                    self.skip_whitespace()
                # handle the hexadecimal character specification case
                elif self.current_char == "x":
                    self.advance()
                    digits: str = self.current_char
                    # needs to have exactly 2 digits
                    if self.current_char not in HEX_NUMBER:
                        self.error("Invalid hex digit", column=column)
                    self.advance()
                    digits += self.current_char
                    if self.current_char not in HEX_NUMBER:
                        self.error("Invalid hex digit", column=column)
                    self.advance()
                    result += self._safe_decode(int(digits, 16))
                elif self.current_char == "u":
                    self.advance()
                    if self.current_char != "{":
                        self.error("Invalid character after \\u, expected {")
                    self.advance()
                    codepoint: str = ""
                    for _ in range(8):
                        codepoint += self.current_char
                        if self.current_char not in HEX_NUMBER:
                            self.error(
                                "Invalid unicode codepoint, expected hexadecimal number"
                            )
                        self.advance()
                        if self.current_char == "}":
                            break
                    else:
                        self.error(
                            "Unicode codepoints can be at most 2^31 - 1, did not close unicode escape"
                        )
                    result += self._safe_code_point(int(codepoint, 16))
                # handle the decimal character specification case
                elif self.current_char in NUMBER:
                    digits = self.current_char
                    self.advance()
                    # may have up to 3 digits
                    if self.current_char in NUMBER:
                        digits += self.current_char
                        self.advance()
                    if self.current_char in NUMBER:
                        digits += self.current_char
                        self.advance()
                    char: int = int(digits)
                    if char > 255:
                        self.error(f"Invalid char with number {char}", column=column)
                    result += self._safe_decode(char)
                else:
                    # handle all simple escape sequences
                    if self.current_char not in ESCAPE_CODES:
                        self.error(
                            f"Invalid escape sequence: \\{self.current_char}",
                            column=column,
                        )
                    result += ESCAPE_CODES[self.current_char]
                    self.advance()
                escape = False
            elif self.current_char == "\\":
                # escape the next character. don't do anything in this loop
                escape = True
                self.advance()
            else:
                # normal characters
                if self.current_char == "\n":
                    self.error("Invalid end of string")
                result += self.current_char
                self.advance()
        # eof before closing character
        if self.current_char != closing:
            self.error("Did not close string")
        self.advance()
        return result

    def get_next_token(self) -> Token:
        # skip prelude
        if self.line == 0 and self.column == 0 and self.current_char == "#":
            while self.current_char and self.current_char != "\n":
                self.advance()
        while self.current_char:
            self.last_hint = None
            # skip whitespace
            if self.current_char.isspace():
                self.skip_whitespace()
                continue

            # skip comments
            if self.current_char == "-" and self.peek() == "-":
                self.skip_comment()
                continue

            args: TokenArgs = self.get_token_args()

            if self.current_char in LETTER:
                name: str = self.get_name()
                if token_type := RESERVED_KEYWORDS.get(name):
                    return Token(token_type, name, **args)
                return Token(TokenType.NAME, name, **args)

            if (
                self.current_char in NUMBER
                or self.current_char == "."
                and self.peek() in NUMBER
            ):
                number: NumberTuple = self.get_number()
                return Token(TokenType.NUMBER, number, **args)

            if self.current_char in ["'", '"']:
                string: str = self.get_string()
                return Token(TokenType.STRING, string, **args)

            if self.current_char == "[" and self.peek() in ["[", "="]:
                string = self.get_long_brackets()
                return Token(TokenType.STRING, string, **args)

            if self.current_char == "." and self.peek() == ".":
                self.advance()
                self.advance()
                if self.current_char == ".":
                    self.advance()
                    return Token(TokenType.ELLIPSIS, "...", **args)
                return Token(TokenType.CONCAT, "..", **args)

            if peek := self.peek():
                double_character: str = self.current_char + peek
                if token_type := SYMBOLS.get(double_character):
                    self.advance()
                    self.advance()
                    return Token(token_type, double_character, **args)
            char: str = self.current_char
            if token_type := SYMBOLS.get(char):
                self.advance()
                return Token(token_type, char, **args)

            self.error(f"unrecognised character {self.current_char}")
        return Token(TokenType.EOF, "eof", **self.get_token_args())



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from .lexer import NumberTuple


class TokenType(Enum):
    AND = "and"
    BREAK = "break"
    DO = "do"
    ELSE = "else"
    ELSEIF = "elseif"
    END = "end"
    FALSE = "false"
    FOR = "for"
    FUNCTION = "function"
    GOTO = "goto"
    IF = "if"
    IN = "in"
    LOCAL = "local"
    NIL = "nil"
    NOT = "not"
    OR = "or"
    REPEAT = "repeat"
    RETURN = "return"
    THEN = "then"
    TRUE = "true"
    UNTIL = "until"
    WHILE = "while"
    AS = "as"
    IS = "is"
    PLUS = "+"
    MINUS = "-"
    MULT = "*"
    DIVIDE = "/"
    MODULO = "%"
    EXPONENT = "^"
    HASH = "#"
    EQUALS = "=="
    NOT_EQUALS = "~="
    LESS_EQUALS = "<="
    GREATER_EQUALS = ">="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    ASSIGN = "="
    L_PAREN = "("
    R_PAREN = ")"
    L_CURL = "{"
    R_CURL = "}"
    L_BRACKET = "["
    R_BRACKET = "]"
    SEMICOLON = ";"
    COLON = ":"
    LABEL_BORDER = "::"
    COMMA = ","
    DOT = "."
    CONCAT = ".."
    ELLIPSIS = "..."
    BIT_AND = "&"
    BIT_OR = "|"
    BIT_XOR = "~"
    BIT_SHIFT_LEFT = "<<"
    BIT_SHIFT_RIGHT = ">>"
    INTEGER_DIVISION = "//"
    NAME = "name"
    NUMBER = "number"
    STRING = "string"
    EOF = "eof"


class Token:
    def __init__(
        self,
        type: TokenType,
        value: Union[str, bool, NumberTuple],
        line: int,
        column: int,
        comment: Optional[list[str]] = None,
    ) -> None:
        self.type: TokenType = type
        self.value: Union[str, bool, NumberTuple] = value
        self.line: int = line
        self.column: int = column
        self.comment: list[str] = comment or []

    def __hash__(self) -> int:
        return hash((self.type, self.value))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Token):
            return self.type == other.type and self.value == other.value
        return False

    def __repr__(self) -> str:
        return f"Token({self.type}, {self.value!r}, {self.line!r}, {self.column!r}, {self.comment!r})"



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import Any, Generator, Generic, Optional, TypeVar

from tumfl.Token import Token
from tumfl.utils import generic_str

T = TypeVar("T")


class ASTNode(ABC, Generic[T]):
    """Base class for all AST nodes"""

    def __init__(self, token: Token, name: str) -> None:
        self.name: str = name
        self.token: Token = token
        self.parent_class: Optional[ASTNode] = None
        self.file_name: Optional[Path] = None
        self.attributes: Optional[T] = None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return all(
            callable(getattr(self, i)) or getattr(self, i) == getattr(other, i)
            for i in self.__dir()
        )

    def __repr__(self) -> str:
        return generic_str(self, ["replace", "parent", "parent_class"])

    def __dir(self) -> Generator[str, None, None]:
        return (
            i
            for i in dir(self)
            if not i.startswith("__")
            # ignore "token" for comparison (and parent check)
            and i
            not in ["replace", "parent", "parent_class", "var", "token", "comment"]
        )

    def parent(
        self, parent: Optional[ASTNode], file_name: Optional[Path] = None
    ) -> None:
        self.parent_class = parent
        self.file_name = file_name
        for i in self.__dir():
            node: Any = getattr(self, i)
            if isinstance(node, ASTNode):
                node.parent(self, file_name)
            elif (
                isinstance(node, list)
                and len(node) > 0
                and isinstance(node[0], ASTNode)
            ):
                for node in node:
                    node.parent(self, file_name)

    def replace_child(self, to_replace: ASTNode, replacement: ASTNode) -> None:
        for i in self.__dir():
            node: Any = getattr(self, i)
            if node is to_replace:
                setattr(self, i, replacement)
                return
            if isinstance(node, list):
                for j, element in enumerate(node):
                    if element is to_replace:
                        node[j] = replacement
                        return



import unittest

from tumfl.lexer import *


class TestNumberLexing(unittest.TestCase):
    def test_simple_int(self):
        lex = Lexer("1")
        nmb: NumberTuple = (False, "1", None, None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("123409204023492")
        nmb: NumberTuple = (False, "123409204023492", None, None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_hex_int(self):
        lex = Lexer("0x123")
        nmb: NumberTuple = (True, "123", None, None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0XaBc123")
        nmb: NumberTuple = (True, "abc123", None, None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_simple_float(self):
        lex = Lexer("0.1")
        nmb: NumberTuple = (False, "0", "1", None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("123.456")
        nmb: NumberTuple = (False, "123", "456", None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_hex_simple_float(self):
        lex = Lexer("0x0.a")
        nmb: NumberTuple = (True, "0", "a", None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0X5.Ca")
        nmb: NumberTuple = (True, "5", "ca", None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_float_dangling_dot(self):
        lex = Lexer(".3")
        nmb: NumberTuple = (False, None, "3", None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0.")
        nmb: NumberTuple = (False, "0", None, None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_hex_float_dangling_dot(self):
        lex = Lexer("0x.3")
        nmb: NumberTuple = (True, None, "3", None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_float_exponent(self):
        lex = Lexer("0.2e+1")
        nmb: NumberTuple = (False, "0", "2", "+1", None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0.5e")
        nmb: NumberTuple = (False, "0", "5", None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("34E1")
        nmb: NumberTuple = (False, "34", None, "1", None)
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0E+1")
        nmb: NumberTuple = (False, "0", None, "+1", None)
        self.assertEqual(lex.get_number(), nmb)

    def test_hex_float_offset(self):
        lex = Lexer("0x2.efp3")
        nmb: NumberTuple = (True, "2", "ef", None, "3")
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0x.4p-2")
        nmb: NumberTuple = (True, None, "4", None, "-2")
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0X1.921FB54442D18P+1")
        nmb: NumberTuple = (True, "1", "921fb54442d18", None, "+1")
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0x.0p-3")
        nmb: NumberTuple = (True, None, "0", None, "-3")
        self.assertEqual(lex.get_number(), nmb)
        lex = Lexer("0xE+1")
        nmb: NumberTuple = (True, "e", None, None, None)
        self.assertEqual(lex.get_number(), nmb)
        lex.advance()
        nmb: NumberTuple = (False, "1", None, None, None)
        self.assertEqual(lex.get_number(), nmb)

    def test_boundaries(self):
        lex = Lexer("0.1a")
        nmb: NumberTuple = (False, "0", "1", None, None)
        self.assertEqual(lex.get_number(), nmb)
        self.assertEqual(lex.current_char, "a")
        lex = Lexer("0xag")
        nmb: NumberTuple = (True, "a", None, None, None)
        self.assertEqual(lex.get_number(), nmb)
        self.assertEqual(lex.current_char, "g")
        lex = Lexer("0.3p2")
        nmb: NumberTuple = (False, "0", "3", None, None)
        self.assertEqual(lex.get_number(), nmb)
        self.assertEqual(lex.current_char, "p")
        lex = Lexer(".3e-5c")
        nmb: NumberTuple = (False, None, "3", "-5", None)
        self.assertEqual(lex.get_number(), nmb)
        self.assertEqual(lex.current_char, "c")
        lex = Lexer("1xb")
        nmb: NumberTuple = (False, "1", None, None, None)
        self.assertEqual(lex.get_number(), nmb)
        self.assertEqual(lex.current_char, "x")

    def test_only_fraction(self):
        lex = Lexer(".4")
        nmb: NumberTuple = (False, None, "4", None, None)
        tok = Token(TokenType.NUMBER, nmb, 0, 0)
        self.assertEqual(lex.get_next_token(), tok)


class TestGetLongBrackets(unittest.TestCase):
    def test_single(self):
        lex = Lexer("[[123456''\\\n]]")
        self.assertEqual(lex.get_long_brackets(), "123456''\\\n")
        self.assertEqual(lex.current_char, None)

    def test_multiple(self):
        lex = Lexer("[=[==[===[1245121[[\\\n]]]==]]=]a")
        self.assertEqual(lex.get_long_brackets(), "==[===[1245121[[\\\n]]]==]")
        self.assertEqual(lex.current_char, "a")

    def test_fails(self):
        lex = Lexer("[[dklaospkda\\")
        with self.assertRaises(LexerError):
            lex.get_long_brackets()
        lex = Lexer("[==[[[]]]=]]===]")
        with self.assertRaises(LexerError):
            lex.get_long_brackets()

    def test_no_long_bracket(self):
        lex = Lexer("[=======")
        with self.assertRaises(LexerError):
            lex.get_long_brackets()


class TestSkipComment(unittest.TestCase):
    def test_single_line_comment(self):
        lex = Lexer("--123456''\n1")
        lex.skip_comment()
        self.assertEqual(lex.current_char, "\n")

    def test_long_comment(self):
        lex = Lexer("--[[abc123\ndklask12''\\\n]]a")
        lex.skip_comment()
        self.assertEqual(lex.current_char, "a")


class TestSkipWhitespace(unittest.TestCase):
    def test_skip_whitespace(self):
        lex = Lexer("  \t\n    \n  \ta")
        lex.skip_whitespace()
        self.assertEqual(lex.current_char, "a")


class TestGetName(unittest.TestCase):
    def test_simple(self):
        lex = Lexer("abcdefg")
        self.assertEqual(lex.get_name(), "abcdefg")
        lex = Lexer("ab1234ad_")
        self.assertEqual(lex.get_name(), "ab1234ad_")
        lex = Lexer("A_Z")
        self.assertEqual(lex.get_name(), "A_Z")

    def test_boundaries(self):
        lex = Lexer("abz123-4")
        self.assertEqual(lex.get_name(), "abz123")


class TestGetString(unittest.TestCase):
    def test_simple(self):
        lex = Lexer('"abc"')
        self.assertEqual(lex.get_string(), "abc")
        lex = Lexer("'abc'")
        self.assertEqual(lex.get_string(), "abc")
        self.assertEqual(lex.current_char, None)

    def test_skip_whitespace(self):
        lex = Lexer("' \\z    \n\ta'")
        self.assertEqual(lex.get_string(), " a")
        lex = Lexer("'\\za'")
        self.assertEqual(lex.get_string(), "a")

    def test_other_quote(self):
        lex = Lexer("'\"'")
        self.assertEqual(lex.get_string(), '"')
        lex = Lexer('"\'"')
        self.assertEqual(lex.get_string(), "'")

    def test_fail_eof(self):
        lex = Lexer("'abc")
        with self.assertRaises(LexerError):
            lex.get_string()

    def test_fail_line_end(self):
        lex = Lexer("'abc\n'")
        with self.assertRaises(LexerError):
            lex.get_string()

    def test_simple_escape_codes(self):
        lex = Lexer(r"'\a\b\f\n\r\t\v\\\"\''")
        self.assertEqual(lex.get_string(), "\a\b\f\n\r\t\v\\\"'")
        lex = Lexer("'\\\n'")
        self.assertEqual(lex.get_string(), "\n")
        with self.assertRaises(LexerError):
            lex = Lexer("'\\A'")
            lex.get_string()
        with self.assertRaises(LexerError):
            lex = Lexer("'\\e'")
            lex.get_string()

    def test_hexadecimal_escape(self):
        lex = Lexer("'\\x1'")
        with self.assertRaises(LexerError):
            lex.get_string()
        lex = Lexer("'\\x'")
        with self.assertRaises(LexerError):
            lex.get_string()
        lex = Lexer("'\\x0a'")
        self.assertEqual(lex.get_string(), "\n")
        lex = Lexer("'\\x61'")
        self.assertEqual(lex.get_string(), "a")

    def test_decimal_escape(self):
        lex = Lexer("'\\0'")
        self.assertEqual(lex.get_string(), "\0")
        lex = Lexer("'\\256'")
        with self.assertRaises(LexerError):
            lex.get_string()
        lex = Lexer("'\\97\\0971'")
        self.assertEqual(lex.get_string(), "aa1")
        lex = Lexer("'\\97lo\\10\\04923\"'")
        self.assertEqual(lex.get_string(), 'alo\n123"')


class TestNextToken(unittest.TestCase):
    def test_simple(self):
        lex = Lexer("a=2.45e-3;d=a+4<55")
        self.assertEqual(lex.get_next_token(), Token(TokenType.NAME, "a", 1, 1))
        self.assertEqual(lex.get_next_token(), Token(TokenType.ASSIGN, "=", 1, 2))
        nmb: NumberTuple = (False, "2", "45", "-3", None)
        self.assertEqual(lex.get_next_token(), Token(TokenType.NUMBER, nmb, 1, 3))
        self.assertEqual(lex.get_next_token(), Token(TokenType.SEMICOLON, ";", 1, 10))
        self.assertEqual(lex.get_next_token(), Token(TokenType.NAME, "d", 1, 11))
        self.assertEqual(lex.get_next_token(), Token(TokenType.ASSIGN, "=", 1, 12))
        self.assertEqual(lex.get_next_token(), Token(TokenType.NAME, "a", 1, 13))
        self.assertEqual(lex.get_next_token(), Token(TokenType.PLUS, "+", 1, 14))
        nmb: NumberTuple = (False, "4", None, None, None)
        self.assertEqual(lex.get_next_token(), Token(TokenType.NUMBER, nmb, 1, 15))
        self.assertEqual(lex.get_next_token(), Token(TokenType.LESS_THAN, "<", 1, 16))
        nmb: NumberTuple = (False, "55", None, None, None)
        self.assertEqual(lex.get_next_token(), Token(TokenType.NUMBER, nmb, 1, 17))
        self.assertEqual(lex.get_next_token(), Token(TokenType.EOF, "eof", 1, 19))
        self.assertEqual(lex.get_next_token(), Token(TokenType.EOF, "eof", 1, 19))

    def test_double_character_token(self):
        lex = Lexer("==")
        self.assertEqual(lex.get_next_token(), Token(TokenType.EQUALS, "==", 0, 0))
        self.assertEqual(lex.get_next_token(), Token(TokenType.EOF, "eof", 0, 0))

    def test_comment_string(self):
        lex = Lexer("'abc\\\ndef'--abc\n\"abab'\"--[==[\n\\\n]===]]==]'abc'")
        self.assertEqual(
            lex.get_next_token(), Token(TokenType.STRING, "abc\ndef", 1, 1)
        )
        self.assertEqual(lex.get_next_token(), Token(TokenType.STRING, "abab'", 2, 1))
        self.assertEqual(lex.get_next_token(), Token(TokenType.STRING, "abc", 5, 10))
        self.assertEqual(lex.get_next_token(), Token(TokenType.EOF, "eof", 5, 15))

    def test_tokenizer_error(self):
        lex = Lexer("!")
        with self.assertRaises(LexerError):
            lex.get_next_token()


class TestNil(unittest.TestCase):
    def test_nil(self):
        lex = Lexer("nil")
        self.assertEqual(lex.get_next_token(), Token(TokenType.NIL, "nil", 1, 1))


class TestComment(unittest.TestCase):
    def test_comment_inclusion(self):
        lex: Lexer = Lexer("--The following line is special\nabc")
        tok: Token = lex.get_next_token()
        self.assertEqual(tok.comment, ["The following line is special"])
        self.assertEqual(tok.line, 2)
        self.assertEqual(tok.column, 1)
        lex: Lexer = Lexer("--some comment\n--another comment\nabc")
        tok: Token = lex.get_next_token()
        self.assertEqual(tok.comment, ["some comment", "another comment"])


class TestEllipsis(unittest.TestCase):
    def test_ellipsis(self):
        lex = Lexer("...")
        self.assertEqual(lex.get_next_token(), Token(TokenType.ELLIPSIS, "...", 0, 0))
        self.assertEqual(lex.get_next_token(), Token(TokenType.EOF, "eof", 0, 0))
        lex = Lexer("..a...b")
        self.assertEqual(lex.get_next_token(), Token(TokenType.CONCAT, "..", 0, 0))
        self.assertEqual(lex.get_next_token(), Token(TokenType.NAME, "a", 0, 0))
        self.assertEqual(lex.get_next_token(), Token(TokenType.ELLIPSIS, "...", 0, 0))
        self.assertEqual(lex.get_next_token(), Token(TokenType.NAME, "b", 0, 0))


class TestShebang(unittest.TestCase):
    def test_shebang(self):
        lex = Lexer("#!abc\na")
        self.assertEqual(lex.get_next_token().type, TokenType.NAME)


class EmptyTest(unittest.TestCase):
    def test_empty(self):
        lex = Lexer("")
        self.assertEqual(lex.get_next_token().type, TokenType.EOF)
